import yaml, json
import luna.transforms

def load(stream):
    if isinstance(stream, str):
        stream = open(stream, 'r')
    pipeline_config = yaml.safe_load( stream )
    pipeline = []
    for stage, job_config in enumerate(pipeline_config['stages']):
        print (f"STAGE {stage}: {job_config['job']}")
        print (json.dumps(job_config, indent=4))
        method_to_call = getattr(luna.transforms, job_config['job'])
        pipeline.append ( (method_to_call, job_config) )

    return pipeline




#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on October 17, 2019

@author: pashaa@mskcc.org
'''
import pytest

from luna.common.config import ConfigSet
from luna.common.utils import get_absolute_path


@pytest.fixture(autouse=True)
def cfg():
    print('------setup------')
    c1 = ConfigSet()
    if c1 is not None:
        c1.clear()

    yield cfg

    print('------teardown------')
    c1 = ConfigSet()
    if c1 is not None:
        c1.clear()


def test_singleton_invocations(cfg):
    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')
    c2 = ConfigSet()

    assert c1 == c2  # instance is reused for the same config.yaml

    assert c1.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 111
    assert c2.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 111


def test_singleton(cfg):
    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')
    c2 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')

    assert c1 == c2  # instance is reused for the same config.yaml

    assert c1.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 111
    assert c2.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 111

    c3 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/another_test_config.yml')
    c4 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/another_test_config.yml')

    # instance is reused for different configs too
    assert c3 == c4
    assert c2 == c3

    # but values are reloaded when different config file is provided for the same logical name
    assert c3.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 222
    assert c4.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 222


def test_parse_path():
    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')

    parsed = c1._parse_path('name::jsonpath')

    assert parsed['name'] == 'name'
    assert parsed['jsonpath'] == 'jsonpath'

    parsed = c1._parse_path('name::jsonpath::jsonpath')

    assert parsed['name'] == 'name'
    assert parsed['jsonpath'] == 'jsonpath::jsonpath'

    with pytest.raises(ValueError) as ve:
        c1._parse_path('name:jsonpath')

    assert "Illegal config path: " in str(ve.value)



def test_get_value(cfg):
    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')
    c3 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')

    assert c1.get_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]') == 111

    with pytest.raises(ValueError):
        c3.get_value(path='app_config::$.spark_application_config[:1]["doesnt.exist"]') == None

    with pytest.raises(ValueError):
        ConfigSet().get_value('name_does_not_exist::$.spark_application_config[:1]["spark.executor.cores"]')


def test_get_keys(cfg):
    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')

    assert c1.get_keys('app_config') == ['spark_cluster_config', 'spark_application_config']

    with pytest.raises(ValueError):
        ConfigSet().get_keys('name_does_not_exist')


def test_singleton_with_schema(cfg):
    schema_file = get_absolute_path(__file__, '../../../../pyluna-core/luna/data_ingestion_template_schema.yml')

    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')
    c2 = ConfigSet(name='data_config', config_file='pyluna-common/tests/luna/common/test_data_ingestion_template.yml',
                   schema_file=schema_file)

    assert c1 == c2

    assert c1.get_names() == ['app_config', 'data_config']
    assert c2.get_names() == ['app_config', 'data_config']


def test_invalid_yaml(cfg):
    with pytest.raises(IOError):
        ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/does_not_exist.yml')


def test_has_value(cfg):
    c1 = ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')

    assert c1.has_value(path='app_config::$.spark_application_config[:1]["spark.executor.cores"]')
    assert not c1.has_value(path='app_config::$.spark_application_config[:1]["spark.does.not.exist"]')

    with pytest.raises(ValueError):
        ConfigSet().has_value('name_does_not_exist::$.spark_application_config[:1]["spark.does.not.exist"]')

def test_clear(cfg):
    schema_file = get_absolute_path(__file__, '../../../../pyluna-core/luna/data_ingestion_template_schema.yml')

    ConfigSet(name='app_config', config_file='pyluna-common/tests/luna/common/test_config.yml')
    ConfigSet(name='data_config', config_file='pyluna-common/tests/luna/common/test_data_ingestion_template.yml',
                   schema_file=schema_file)
    c1 = ConfigSet()
    assert c1.get_names() == ['app_config', 'data_config']
    c1.clear()
    assert c1.get_names() == []



from neo4j import GraphDatabase
from neo4j import __version__ as neo4j_version

from pyspark.sql.types import StringType,StructType,StructField
import re

def pretty_path(path):
    to_print = ''
    for i,x in enumerate(path):
        if type(x)==dict:
            node_desc = '(' + ','.join([key+":"+x[key] for key in x.keys()]) + ')'
            if   i==0: 		   to_print += "SOURCE:" + node_desc # First
            elif i==(len(path)-1): to_print += "SINK:" + node_desc # Last
            else:                  to_print += node_desc # Middle
        if type(x)==str:
            to_print += '-[' + x + ']-'
    return to_print

class Neo4jConnection:

    def __init__(self, uri, user, pwd):
        self.__uri = uri
        self.__user = user
        self.__pwd = pwd
        self.__driver = None
        try:
            self.__driver = GraphDatabase.driver(self.__uri, auth=(self.__user, self.__pwd))
        except Exception as ex:
            print("Failed to create the driver: ", ex)

    def close(self):
        if self.__driver is not None:
            self.__driver.close()

    def query(self, query, db=None):
        """
        Runs a cyper query against the initalized driver
        """
        assert self.__driver is not None, "Driver not initialized!"
        session = None
        response = None
        try:
            session = self.__driver.session(database=db) if db is not None else self.__driver.session()
            response = list(session.run(query))
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()
        return response

    def test_connection(self):
        try:
            self.__driver.session().run("MATCH () RETURN 1 LIMIT 1")
            return True
        except Exception:
            return False



    # ==========================================================================================================
    # Simple methods
    def test_count(self):
        """
        Get node(s) given a value input
        """
        result = self.query(f"""
            MATCH (n) RETURN n
        """
        )
        node_count = len(result)
        print (f"Successfully connected to {node_count} nodes!")

    def match_concept_node(self, QUERY_ID):
        """
        Get node(s) given a value input
        """
        result = self.query(f"""
            MATCH (concept_node)
            WHERE concept_node.value = '{QUERY_ID}'
            RETURN concept_node"""
        )
        return [(x.data(),) for x in result]


    # ==========================================================================================================
    # This should become the main helper method
    # Leaving others in for now for backwards compatability
    def create_id_lookup_table_where(self, sqlc, source, sink, r="ID_LINK|HAS_RECORD", WHERE_CLAUSE=""):
        """
        Main method for creating lookup tables in a general manner.

        Args:
            source: what ID type you are starting with (e.g. a dmp_patient_id)
		    sink: what ID you wish to map to (e.g. a SeriesInstanceUID)
            r: optional. Allowed relationship types.  Default is ID_LINK|HAS_RECORD to contain mapping within a patient subgraph
		    WHERE_CLAUSE: optional. source, sink (as nodes) and r (as relationships) become availabe for filtering in a WHERE clause

        Returns:
             a dataframe with three columns, [ source | sink | pathspec ] where pathspec is a text-based representation of the path between the source and sink IDs

        """

        banned_words = ['delete', 'merge', 'create', 'set', 'remove']
        if re.compile('|'.join(banned_words),re.IGNORECASE).search(WHERE_CLAUSE): #re.IGNORECASE is used to ignore case
            raise Exception("You tried to alter the database, goodbye")
            return

        print (f""">>> QUERY >>> \n\tMATCH (source:{source})-[r:{r}*]-(sink:{sink}), \n\tpath=shortestPath( (source)-[:{r}*..15]-(sink) ) \n\t{WHERE_CLAUSE} RETURN DISTINCT source,sink,path""")

        result = self.query(f"""
            MATCH (source:{source})-[r:{r}*]-(sink:{sink}), path=shortestPath( (source)-[:{r}*..15]-(sink) ) \
            {WHERE_CLAUSE} \
            RETURN DISTINCT source,sink,path
            """
        )
        if result is None:
            print ("Improper query returning null")
            return None
        cSchema = StructType([StructField(source, StringType(), True), StructField(sink, StringType(), True), StructField("pathspec", StringType(), True)])
        return sqlc.createDataFrame(([(x.data()['source']['value'],x.data()['sink']['value'], pretty_path(x.data()['path'])) for x in result]),schema=cSchema)
    # ==========================================================================================================


    # >>>>>>>> To depreciate over time <<<<<<<<
    def commute_source_id_to_spark(self, sc, sqlc, SOURCE_TYPE, SINK_TYPE, QUERY_ID ):
        """
        Spark connector for an input source id
        Returns a dataframe with one column named the sink/target ID
        """
        result = self.query(f"""
            MATCH (source:{SOURCE_TYPE})-[:ID_LINK*]-(sink:{SINK_TYPE}) \
            WHERE source.value = '{QUERY_ID}' \
            RETURN source,sink """
        )
        cSchema = StructType([StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(sc.parallelize([(x.data()['sink']['value'],) for x in result]),schema=cSchema)

    def commute_source_id_to_spark_query(self, sc, sqlc, WHERE_CLAUSE, SINK_TYPE ):
        """
        Spark connector for an input source id
        Returns a dataframe with one column named the sink/target ID
        """
        result = self.query(f"""
            MATCH path=(source)-[r*]-(sink:{SINK_TYPE}) \
            {WHERE_CLAUSE} \
            RETURN source,sink,path """
        )
        for x in result: print (pretty_path(x.data()['path']))
        cSchema = StructType([StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(([(x.data()['sink']['value'],) for x in result]),schema=cSchema)
    def create_id_lookup_table(self, sc, sqlc, SOURCE_TYPE, SINK_TYPE, QUERY_ID ):
        """
        Spark connector for an input source id
        Returns a dataframe with two ID columns as specified by SOURCE_TYPE and SINK_TYPE
        """
        result = self.query(f"""
            MATCH (source:{SOURCE_TYPE})-[:ID_LINK|HAS_RECORD*]-(sink:{SINK_TYPE}) \
            WHERE source.value = '{QUERY_ID}' \
            RETURN source,sink """
        )
        cSchema = StructType([StructField(SOURCE_TYPE, StringType(), True), StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(sc.parallelize([(x.data()['source']['value'],x.data()['sink']['value']) for x in result]),schema=cSchema)

    def commute_cohort_id_to_spark (self, sc, sqlc, SOURCE_TYPE, SINK_TYPE, QUERY_ID):
        """
        Spark connector for an input cohort id
        Returns a dataframe with one column named the sink/target ID
        """
        result = self.query(f"""
            MATCH (source:{SOURCE_TYPE})-[:COHORT_LINK*1..1]-(entry_node)
            WHERE source.value = '{QUERY_ID}' \
            MATCH (entry_node)-[:ID_LINK*]-(sink:{SINK_TYPE})
            RETURN source,sink """
        )
        cSchema = StructType([StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(sc.parallelize([(x.data()['sink']['value'],) for x in result]),schema=cSchema)

    def commute_record_id_to_spark (self, sc, sqlc, SOURCE_TYPE, SINK_TYPE):
        """
        Spark connector for an input cohort id
        Returns a dataframe with one column named the sink/target ID
        """
        result = self.query(f"""
            MATCH (source:{SOURCE_TYPE})-[:HAS_RECORD*1..1]-(sink:{SINK_TYPE})
            RETURN source,sink """
        )
        cSchema = StructType([StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(sc.parallelize([(x.data()['sink']['value'],) for x in result]),schema=cSchema)

    def commute_sink_id_to_spark (self, sc, sqlc, SINK_TYPE, QUERY_ID):
        """
        A pass through function for consistency
        """
        cSchema = StructType([StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(sc.parallelize([(x,) for x in [QUERY_ID]]),schema=cSchema)

    def commute_all_sink_id(self, sc, sqlc, SINK_TYPE):
        """
        Spark connector for an input source id
        Returns a dataframe with one column named the sink/target ID
        """
        result = self.query(f"""
            MATCH (sink:{SINK_TYPE}) \
            RETURN sink """
        )
        cSchema = StructType([StructField(SINK_TYPE, StringType(), True)])
        return sqlc.createDataFrame(sc.parallelize([(x.data()['sink']['value'],) for x in result]),schema=cSchema)



from luna.common.custom_logger import init_logger
from testfixtures import LogCapture
import pytest
import os, logging

@pytest.fixture()
def logger():
    os.environ['LUNA_HOME'] = 'pyluna-common/tests/luna/common/testdata/'
    logger = init_logger()
    yield logger

def test_no_luna_home():
    os.environ['LUNA_HOME'] = ''
    with pytest.raises(RuntimeError):
        init_logger()

def test_check_level(logger):
    # 20 for info. ref: https://docs.python.org/3/library/logging.html#levels
    assert 20 == logger.getEffectiveLevel()

def test_levels(logger):
    with LogCapture() as l:
        logger.info('a message')
        logger.warning('a warning')
        logger.error('an error')
        l.check(
            ('root', 'INFO', 'a message'),
            ('root', 'WARNING', 'a warning'),
            ('root', 'ERROR', 'an error'),
        )

def test_formatting(logger):
    with LogCapture() as l:
        logger.info("a message with string %s", "some output")
        l.check(('root', 'INFO', 'a message with string some output'))

def test_list(logger):
    with LogCapture() as l:
        logger.info([1,2,3])
        l.check(('root', 'INFO', '[1, 2, 3]'))

def test_dict(logger):
    with LogCapture() as l:
        logger.info({'list':[1,2,3], 'name':'foo'})
        l.check(('root', 'INFO', "{'list': [1, 2, 3], 'name': 'foo'}"))

# For some reason, this capturing no longer works
# def test_newlines(logger, caplog):
#     logger.info("some\nmultiline\noutput")
#     out = caplog.text
#     print ("out, err=", out)
#     # Make sure multi-lines appear with their own header, make sure to run with pytest -s
#     assert "root - INFO - output" in err



import os
import logging
from logging.handlers import RotatingFileHandler
from log4mongo.handlers import MongoHandler

from luna.common.config import ConfigSet


class MultilineFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord):
        save_msg = str(record.msg)
        output = ""
        for line in save_msg.splitlines():
            record.msg = line
            output += super().format(record) + "\n"
        output = output.rstrip()
        record.msg = save_msg
        record.message = output
        return output


def init_logger(filename="data-processing.log"):
    # Logging configuration
    if os.environ["LUNA_HOME"]:
        cfg = ConfigSet(
            name="LOG_CFG",
            config_file=os.path.join(os.environ["LUNA_HOME"], "conf", "logging.cfg"),
        )
    else:
        raise RuntimeError(
            "$LUNA_HOME is not set. Make sure you have set $LUNA_HOME and $LUNA_HOME/conf/logging.cfg"
        )

    log_file = filename
    logger = logging.getLogger()
    logger.setLevel(cfg.get_value("LOG_CFG::LOG_LEVEL"))
    formatter = MultilineFormatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
    )

    if not logger.handlers:
        # create console handler with a customizable, higher log level
        ch = logging.StreamHandler()
        ch.setLevel(cfg.get_value("LOG_CFG::LOG_LEVEL"))
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        # create file handler which logs even debug messages
        fh = RotatingFileHandler(log_file, maxBytes=1e7, backupCount=10)
        fh.setLevel(cfg.get_value("LOG_CFG::LOG_LEVEL"))
        fh.setFormatter(formatter)
        logger.addHandler(fh)

        # create mongo log handler if we configured it
        if cfg and cfg.get_value("LOG_CFG::CENTRAL_LOGGING"):
            mh = MongoHandler(
                host=cfg.get_value("LOG_CFG::MONGO_HOST"),
                port=cfg.get_value("LOG_CFG::MONGO_PORT"),
                capped=True,
            )
            mh.setLevel(cfg.get_value("LOG_CFG::MONGO_LOG_LEVEL"))
            logger.addHandler(mh)

    logger.info("Initalized logger, log file at: " + log_file)
    logger.debug("Initalized logger with handlers: " + str(logger.handlers))
    return logger



import os
from pathlib import Path
from urllib.parse import urlparse
from functools import partial

import shutil
from minio import Minio
# from minio.error import NoSuchKey
import calendar
from datetime import datetime


def get_object_stats(client, bucket, key):
    try:
        stat_object = client.stat_object(bucket, key)
    except:
        return -1, -1
    return stat_object.size, calendar.timegm(stat_object.last_modified.timetuple())

def get_file_stats(path):
    try:
        stat_file   = os.stat(path)
    except FileNotFoundError:
        return -1, -1
    return stat_file.st_size, stat_file.st_mtime

class WriteAdapter:
    def __init__(self): pass

class NoWriteAdapter(WriteAdapter):
    def __init__(self):
        self.base_url = "file://"
        print("Configured NoWriteAdatper with base URL=" + self.base_url)

    def write(self, input_data, prefix) -> dict:
        """ Returns input_data as written data 
        
        Args:
            input_data (str): path to input file
            prefix (str): relative path prefix for destination, ignored
        Returns:
            dict: key-value pairs containing metadata about the write operation
        """
        input_path  = Path(input_data)
        filename    = input_path.name

        if not os.path.exists(input_path):
            return {}
        
        output_data_url = f"{self.base_url}{input_path}"
        
        input_size,  input_mtime   = get_file_stats  (input_data)
    
        return {'readable': True, 'data_url': output_data_url, 'size': input_size, 'ingest_time': datetime.fromtimestamp(input_mtime)}


class FileWriteAdatper(WriteAdapter):
    def __init__(self, store_url, bucket):
        """ Return a WriteAdapter for a given file I/O scheme and URL
        
        Args:
            store_url (str): root URL for the storage location (e.g. s3://localhost:9000 or file:///data)
            bucket (str): the "bucket" or "parent folder" for the storage location
        """
        url_result = urlparse(store_url)
        
        # All we need is the base path
        self.store_path = url_result.path
        self.bucket = bucket

        # Define base URL
        self.base_url = f'file://{Path(self.store_path)}'
        print("Configured FileAdatper with base URL=" + self.base_url)
        
        print ("Available buckets:")
        for x in os.listdir(self.store_path): print (f' - {x}')

    def write(self, input_data, prefix) -> dict:
        """ Perform write operation to a posix file system
        
        Will not perform write if :
            the content length matches (full copy) and the input modification time is earlier than the ingest time (with a 1 min. grace period)
        
        Args:
            input_data (str): path to input file
            prefix (str): relative path prefix for destination
        Returns:
            dict: key-value pairs containing metadata about the write operation
        """
        input_path  = Path(input_data)
        prefix_path = Path(prefix)
        filename    = input_path.name

        if not os.path.exists(input_path):
            return {}
        
        if prefix_path.is_absolute(): raise RuntimeError("Prefix paths must be relative!") # User needs to know prefixes are relative paths

        output_data_url = os.path.join(self.base_url, self.bucket, prefix, filename)
        
        output_dir = os.path.join(self.store_path, self.bucket, prefix)
        output_data = os.path.join(output_dir, filename)
        os.makedirs(output_dir, exist_ok=True)

        input_size,  input_mtime   = get_file_stats  (input_data)
        output_size, output_mtime  = get_file_stats  (output_data)

        needs_write = ( output_size != input_size or (input_mtime - 60) > output_mtime ) # 60 second grace period

        if needs_write: 
            shutil.copy(input_data, output_data)
        
        output_size, output_mtime  = get_file_stats  (output_data)

        if not output_size> 0: return {'readable': False}

        return {'readable': True, 'data_url': output_data_url, 'size': output_size, 'ingest_time': datetime.fromtimestamp(output_mtime)}
        
class MinioWriteAdatper(WriteAdapter):
    def __init__(self, store_url, bucket, secure=False):
        """ Return a WriteAdapter for a given file I/O scheme and URL
        
        Args:
            store_url (str): root URL for the storage location (e.g. s3://localhost:9000 or file:///data)
            bucket (str): the "bucket" or "parent folder" for the storage location
        """       

        url_result = urlparse(store_url)

        # We need a bit more detail here
        self.store_hostname = url_result.hostname
        self.store_port = url_result.port
        self.bucket = bucket
        self.secure = secure

        self.client_init = partial(Minio, f'{self.store_hostname}:{self.store_port}', access_key=os.environ['MINIO_USER'], secret_key=os.environ['MINIO_PASSWORD'], secure=secure)

        self.base_url = os.path.join(f"s3://{self.store_hostname}:{self.store_port}")
        print("Configured MinioAdatper with base URL=" + self.base_url)
        
        client = self.client_init()
        
        print ("Available buckets:")
        for x in client.list_buckets(): print (f' - {x.name}')

    def write(self, input_data, prefix) -> dict:
        """ Perform write operation to a s3 file system
        
        Will not perform write if :
            the content length matches (full copy) and the input modification time is earlier than the ingest time (with a 1 min. grace period)

        Args:
            input_data (str): path to input file
            prefix (str): relative path prefix for destination
        Returns:
            dict: key-value pairs containing metadata about the write operation
        """
        input_path  = Path(input_data)
        prefix_path = Path(prefix)
        filename    = input_path.name

        if not os.path.exists(input_path):
            return {}
        
        if prefix_path.is_absolute(): raise RuntimeError("Prefix paths must be relative!") # User needs to know prefixes are relative paths

        output_data_url = os.path.join(self.base_url, self.bucket, prefix, filename)

        client = self.client_init()

        object_size, object_mtime = get_object_stats(client, self.bucket, f"{prefix_path}/{filename}")
        input_size,  input_mtime  = get_file_stats  (input_data)
        needs_write = ( object_size != input_size or (input_mtime - 60) > object_mtime ) # 60 second grace period

        if needs_write: 
            client.fput_object(self.bucket, f"{prefix_path}/{filename}", input_path, part_size=250000000)
        
        object_size, object_mtime = get_object_stats(client, self.bucket, f"{prefix_path}/{filename}")

        if not object_size > 0: return {'readable': False}

        return {'readable': True, 'data_url': output_data_url, 'size': object_size, 'ingest_time': datetime.fromtimestamp(object_mtime)}
    

class IOAdapter:
    """ Interface for IO 

    Exposes a write and read method via scheme specific classes:

    IOAdapter.writer(<scheme>).write(<data>) -> url

    IOAdapter.reader(<scheme>).read(<url>)   -> data (TODO)
    """
    def __init__(self, no_write=False):
        self.no_write = no_write
    
    def writer(self, store_url, bucket):
        """ Return a WriteAdapter for a given file I/O scheme and URL
        
        Args:
            store_url (str): root URL for the storage location (e.g. s3://localhost:9000 or file:///data)
            bucket (str): the "bucket" or "parent folder" for the storage location
            
        Returns
            WriteAdapter: object capable of writing to the location at store_url
            
        """
        if self.no_write: return NoWriteAdapter()

        url_result = urlparse(store_url)
        
        if url_result.scheme == 'file':
            return FileWriteAdatper(store_url=store_url, bucket=bucket)
        elif url_result.scheme == 's3':
            return MinioWriteAdatper(store_url=store_url, bucket=bucket)
        elif url_result.scheme == 's3+https':
            return MinioWriteAdatper(store_url=store_url, bucket=bucket, secure=True)
        else:
            raise RuntimeError("Unsupported slide store schemes, please try s3:// or file://")



'''
Created on November 16, 2020

@author: rosed2@mskcc.org
'''
import os
# Application Constants

#### Table Names ####
from luna.common.utils import get_absolute_path

TABLE_DIR			='tables/'
# clinical
DIAGNOSIS_TABLE		=TABLE_DIR + 'diagnosis'
MEDICATION_TABLE	=TABLE_DIR + 'medication'
PATIENT_TABLE		=TABLE_DIR + 'patient'

# radiology
DICOM_TABLE			=TABLE_DIR + 'dicom'
SCAN_TABLE 			=TABLE_DIR + 'scan'
SCAN_ANNOTATION_TABLE		=TABLE_DIR + 'scan_annotation'
FEATURE_TABLE		=TABLE_DIR + 'feature'

#### Raw Data Directories ####
DICOMS				='dicoms'
SCANS				='scans'
SCAN_ANNOTATIONS	='scan_annotations'
PATHOLOGY_ANNOTATIONS	='pathology_annotations'
FEATURES			='features'

# Configurations
APP_CFG				='APP_CFG'
DATA_CFG			='DATA_CFG'
SCHEMA_FILE			=get_absolute_path(__file__, '../data_ingestion_template_schema.yml')
PUBLIC_DIR          ='/gpfs/mskmind_ess/mind_public'

ANNOTATION_TABLE_MAPPINGS = {"regional":\
 {"DATA_TYPE":"REGIONAL_METADATA_RESULTS", "GEOJSON_COLUMN_NAME": "geojson"}, \
 "point": {"DATA_TYPE": "POINT_GEOJSON", "GEOJSON_COLUMN_NAME": "geojson"} \
    }



def PROJECT_LOCATION(cfg):
    """
    ROOT_PATH is a path to mind data e.g. /gpfs/mind/data or hdfs://server:port/data

    :param cfg:
    :return: ROOT_PATH/PROJECT_NAME
    """
    # This function assumes <uri_root>/data as "ROOT_PATH"
    return os.path.join(cfg.get_value(path=DATA_CFG+'::ROOT_PATH'), cfg.get_value(path=DATA_CFG+'::PROJECT'))

def CONFIG_LOCATION(cfg):
    return "{0}/configs/{1}".format(PROJECT_LOCATION(cfg), TABLE_NAME(cfg))

def TABLE_LOCATION(cfg, is_source=False):
    return "{0}/tables/{1}".format(PROJECT_LOCATION(cfg), TABLE_NAME(cfg, is_source))

def TABLE_NAME(cfg, is_source=False):

    if is_source:
        table_name = cfg.get_value(path=DATA_CFG+'::SOURCE_DATA_TYPE').upper()
    else:
        table_name = cfg.get_value(path=DATA_CFG+'::DATA_TYPE').upper()

    dataset_name = cfg.get_value(path=DATA_CFG+'::DATASET_NAME')

    if dataset_name != "" and dataset_name is not None:
        table_name += "_{0}".format(dataset_name)

    return table_name



import os
import logging
import dask
import asyncio

from dask.distributed import Client
from dask.distributed import worker_client, get_client, get_worker
from functools import partial

logger = logging.getLogger(__name__)

LENGTH_MANY_TASKS = 50000 # When should we warn the user they are possibly paying a high price?

def prune_empty_delayed(tasks):
    """
    A less-than-ideal method to prune empty tasks from dask tasks 
    Here we're trading CPU and time for memory.
    
    Args:
        tasks (list): list of delayed dask tasks
    
    Returns:
        list[dask.delayed]: a reduced list of delayed dask tasks
    """
    @dask.delayed
    def mock_return(task):
        return task

    if len(tasks) >= LENGTH_MANY_TASKS:
        logger.warning(f"Hope you're okay with a length={len(tasks)} for loop!")

    return [mock_return(task) for task in dask.compute(*tasks) if task is not None]

def get_or_create_dask_client():
    try:
        client = get_client()
    except ValueError:
        client = Client(threads_per_worker=1)
    return client

def get_local_dask_directory():
    local_directory = dask.config.get("temporary-directory") or os.getcwd()
    local_directory = os.path.join(local_directory, "dask-worker-space")
    return local_directory

class JobExecutionError(Exception): pass

def with_event_loop(func):
    """
    This method decorates functions run on dask workers with an async function call
    Namely, this allows us to manage the execution of a function a bit better, and especially, to exit job execution if things take too long (1hr)

    Here, the function func is run in a background thread, and has access to the dask schedular through the 'runner'.
    Critically, sumbission to this runner/client looks the same regardless of if it occurs in a sub-process/thread

    Mostly, this is a workaround to impliment some form of timeout when running very long-tasks on dask. 
    While one cannot (or should not) kill the running thread, Dask will cleanup the child tasks eventually once all jobs finish.

    Examples:
        >>> @with_dask_event_loop
        >>> my_job(args, kwargs, runner=None):
        >>>     runner.submit(sleep, 10)
    """

    async def wrapped(*args, **kwargs):
        loop = asyncio.get_event_loop()        
        
        # Get our current dask worker, functions wrapped with this method can only be run on dask workers
        logger.info ("Initializing job... getting parent worker")
        try:
            worker = get_worker()
        except ValueError as exc:
            logger.error("Could not get dask worker!")
            raise RuntimeError("Data-processing job called without parent dask worker")
        except Exception as exc:
            logger.exception(f"Unknown exception when getting dask worker")

        logger.info (f"Successfully found worker {worker}")
        logger.info (f"Running job {func} with args: {args}, kwargs: {kwargs}")

        # Get our worker client, and pass as a dask client exector
        with worker_client() as runner:

            # We'll run our function in a background thread
            # executor = ProcessPoolExecutor(max_workers=1)

            # Add our runner to kwargs
            kwargs['runner'] = runner

            # Kick off the job
            job = loop.run_in_executor(worker.executor, partial(func, *args, **kwargs))
        
            # Move on from job if things take more than hour
            done, pending = await asyncio.wait([job], timeout=3600)

            # Do some cleanup
            if len(pending) != 0:
                logger.warning ("Killing pending tasks!")
                for task in pending: task.cancel()

            # executor.shutdown(wait=False)

            # Get the return value
            if len(done) == 1:
                return_value = done.pop().result()
            else:
                return_value = None

            # Logg that we're done!
            logger.info (f"Done running job, returning {return_value}")

        return return_value
    
    def run_loop(*args, **kwargs):
        """
        Uses async and threading capabilities

        Use of background thread causes this error on shutdown: 
            ERROR - asyncio - task: <Task pending coro=<HTTP1ServerConnection._server_request_loop() running at /gpfs/mskmindhdp_emc/sw/env/lib64/python3.6/site-packages/tornado/http1connection.py:817> wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x7f52e8259318>()]> cb=[IOLoop.add_future.<locals>.<lambda>() at /gpfs/mskmindhdp_emc/sw/env/lib64/python3.6/site-packages/tornado/ioloop.py:690]>
            Seems like some async task gets hung up in the child thread...
        """

        loop   = asyncio.new_event_loop()
        result = loop.run_until_complete(wrapped(*args, **kwargs))
        loop.close()
        return result
    
def dask_job(job_name):
    """
    The simplier version of a dask job decorator, which only provides the worker_client as a runner to the calling function

    Examples:
        >>> @dask_job('my_job')
        >>> my_job(args, kwargs, runner=None):
        >>>     runner.submit(sleep, 10)
    """

    def wrapped(func):

        def run_simple(namespace, index, *args, **kwargs):
            """
            Only provides runner object to method, no threading
            """
            # Tell us we are running
            logger.info (f"Initializing {job_name} @ {namespace}/{index}")

            # See if we are on a dask worker
            try:
                worker = get_worker()
            except ValueError as exc:
                logger.warning("Could not get dask worker!")
                worker = None
            except Exception as exc:
                logger.exception(f"Unknown exception when getting dask worker")
                worker = None

            logger.info (f"Successfully found worker {worker}")

            # Jobs get an output directory and an ouput parquet slice
            output_dir = os.path.join(os.environ['MIND_GPFS_DIR'], "data_dev", namespace, index)
            output_ds  = os.path.join(os.environ['MIND_GPFS_DIR'], "data_dev", namespace, job_name.upper() + "_DS")
            os.makedirs(output_dir, exist_ok=True)
            os.makedirs(output_ds,  exist_ok=True)
            output_segment = os.path.join(output_ds, f"ResultSegment-{index}.parquet")

            logger.info(f"Setup ouput dir={output_dir} slice={output_ds}")

            # Kick off the job
            try:
                logger.info (f"Running job {func} with args: {args}, kwargs: {kwargs}")
                return_value = func (index, output_dir, output_segment, *args, **kwargs )
            except Exception as exc:
                logger.exception(f"Job execution failed due to: {exc}", extra={ "namespace":namespace, "key": index})
                raise JobExecutionError(f"Job {func} did not run successfully, please check input data! {args}, {kwargs}")

            return return_value

        run_simple.__name__ = job_name

        return run_simple

    return wrapped
